(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gonttypes.ml,v 1.20 2001/11/29 09:55:52 malekith Exp $
 *
 * Module implementation: Gonttypes
 * Description: The typechecker. Term unification is used to check whatever
 * given two types are compatibile. If we use any kind of subtyping this
 * will gonna have to be directed somewhat...
 *)

open Gontutil
open Gontbind

exception Type_error of string

let sorry s = raise (Type_error s)
;;

let rec string_of_type_list tt =
  String.concat ", " (List.map string_of_type tt)
and string_of_type_name_list l =
  (String.concat "" 
        (List.map (fun (t, n) -> (string_of_type t) ^ " " ^ n ^ "; ") l))
and string_of_fv l =
  match l with
      [] -> ""
    | _ -> "<" ^ (String.concat ", " (List.map (fun s -> "'" ^ s) l)) ^ ">" 
and string_of_type t =
  match t with
      Void -> "void"
    | Int -> "int"
    | Float -> "float"
    | Bool -> "bool"
    | String -> "string"
    | Func f -> 
        (string_of_fv f.ft_tyvars) ^ 
        "*(" ^ (string_of_type_list f.ft_parms) ^ ") -> " ^ 
        (string_of_type f.ft_ret_type)
    | TyVar x -> ("'" ^ x)
    | Tuple l ->
        "*[" ^ (string_of_type_list l) ^ "]"
    | Struct l ->
        "struct { " ^ (string_of_type_name_list l) ^ "}"
    | OptStruct l ->
        "opt_struct { " ^ (string_of_type_name_list l) ^ "}"
    | Union l ->
        "union { " ^ (string_of_type_name_list l) ^ "}"
    | AbstractType -> "abstract_type"
    | NamedType n ->
        (match n.nt_parms with
            [] -> ""
          | l -> "<" ^ (String.concat ", " (List.map string_of_type l)) 
                     ^ ">") ^
        (if n.nt_module = "" then
           n.nt_name
         else
           n.nt_module ^ "::" ^ n.nt_name) 
        (* DEBUG *)
        ^
        (match n.nt_type with
            None -> ""
          | Some x -> "/* " ^ string_of_type x ^ " */")
        (* END DEBUG *)
;;

(* o[x/t] *)
let rec subst (x, tt) o =
  let f t =
    match t with
          TyVar y when x = y -> tt
        | TyVar _ | NamedType _ | Int | Void | Bool | Float | String 
        | AbstractType | Tuple _ | Func _ | Struct _ | OptStruct _ 
        | Union _ -> t in
  type_walk f o
;;

exception Is_FV

let is_fv v t =
  let f t =
     match t with
        TyVar x when v = x -> raise Is_FV
      | TyVar _ | Int | Void | Bool | Float | String
      | Func _ | Tuple _ | Union _ | Struct _ | OptStruct _ 
      | NamedType _ | AbstractType -> t in
  try
    ignore(type_walk f t);
    false
  with Is_FV -> true
;;

let rec instance_value tt =
  let f tt =
    match tt with
        NamedType ({nt_type = None; 
                    nt_tydef = Some tydef; 
                    nt_parms = p} as n) ->
          (try
            let mapping = List.map2 (fun x y -> (x, y)) tydef.td_tyvars p in
            let t' = List.fold_left (fun o (x, t) -> subst (x, t) o) 
                                    tydef.td_type 
                                    mapping in
            n.nt_type <- Some t'; t'
          with Invalid_argument _ ->
            err (n.nt_loc, xf "type has %d parameters, but it was instanced \
                               with %d arguments" 
                              (List.length tydef.td_tyvars) 
                              (List.length p)))
      | NamedType {nt_tydef = None} -> assert false 
      | NamedType {nt_type = Some t} -> t
      | t -> t in
   type_walk f tt
;;

(* Term unification, the standard, given-from-God, algorithm :^) *)
let rec mgu (o : (id * ty) list) rrs =
  match rrs with
      r::rs ->
          (match r with
              (TyVar(x), TyVar(y)) when x = y -> 
                  mgu o rs
            | (TyVar(x), t) 
            | (t, TyVar(x)) when not (is_fv x t) ->
                  mgu ((x, t) :: (List.map (fun (x, tt) -> 
                                                (x, subst (x, t) tt)) o))
                      (List.map (fun (t1, t2) -> (subst (x, t) t1, 
                                                  subst (x, t) t2)) rs)
            | (Int, Int) | (Bool, Bool) | (Void, Void)
            | (Float, Float) | (String, String) -> 
                  mgu o rs
            | ((Func {ft_ret_type=r1; ft_parms=p1} as x), 
               (Func {ft_ret_type=r2; ft_parms=p2} as y)) ->
                  if List.length p1 != List.length p2 then
                    sorry (xf "cannot unify function types %s and %s \
                               with diffrent number of parameters"
                              (string_of_type x)
                              (string_of_type y))
                  else
                    mgu o ((r1, r2) :: 
                           (List.map2 (fun a b -> (a, b)) p1 p2) @ rs)
            | ((Tuple l1 as x), (Tuple l2 as y)) ->
                if List.length l1 != List.length l2 then
                  sorry (xf "cannot unfiy tuple types %s and %s of diffrent \
                             arity" 
                            (string_of_type x)
                            (string_of_type y))
                else
                  mgu o ((List.map2 (fun a b -> (a, b)) l1 l2) @ rs)
            | ((Struct(l1) as x), (Struct(l2) as y)) 
            | ((Union(l1) as x), (Union(l2) as y)) 
            | ((OptStruct(l1) as x), (OptStruct(l2) as y)) ->
                (try
                  List.iter2 (fun (t1, n1) (t2, n2) -> 
                     if n1 <> n2 then raise (Invalid_argument "cmp"))
                  l1 l2
                 with Invalid_argument _ -> 
                        sorry (xf "cannot unify %s with %s, since they \
                                   have diffrent struct/union fields"
                                   (string_of_type x)
                                   (string_of_type y)));
                  mgu o ((List.map2 (fun (t1, n1) (t2, n2) -> 
                                                    (t1, t2)) l1 l2) @ rs)
            | (((NamedType _) as x), y) 
            | (x, ((NamedType _) as y)) ->
                  mgu o ((instance_value x, instance_value y) :: rs)
            | (x, y) ->
                    sorry (xf "cannot unify %s with %s" 
                              (string_of_type x)
                              (string_of_type y)))
                                     
    | [] -> o
;;

type ty_context = {ctx_ret_type: ty;   (* Return type of function. *)
                   ctx_tyvars: id list (* Variables functions is parametrized
                                        * over. *)
                  }

  let fresh_vars t =
    let f t =
      match t with
            Func {ft_tyvars = ((_::_) as tv)} -> 
              List.fold_left (fun t x -> subst (x, 
                                                TyVar(seq_name "$")) t) 
                             t tv
          | Func _ | TyVar _ | NamedType _ | Int | Void | Bool | Float | String 
          | AbstractType | Tuple _ | Struct _ | OptStruct _ 
          | Union _ -> t in
  type_walk f t
  
let checker c s t =
  let check_if_nok (v, t) =
    if List.mem v c.ctx_tyvars then
      match t with
          TyVar x when not (List.mem x c.ctx_tyvars) -> (x, TyVar v) (* ok *)
        | _ -> sorry (xf "'%s has to be %s, while it should be \
                          assumption-free" v (string_of_type t))
    else (v, t) in
  let m = mgu [] [(t, s)] in
  List.map check_if_nok m

(* Check if t < s. *)
let subtype c s t =
  ignore(checker c (fresh_vars t) (fresh_vars s))
  
let global_subtype s t =
  subtype {ctx_ret_type = AbstractType; ctx_tyvars = []} s t
  
let typecheck_call c f at =
  let ft = 
    match f with
        Func {ft_parms = x} -> x
      | _ -> sorry (xf "called object is not a function \
                        (it has non-functional type %s)"
                               (string_of_type f)) in
  if (List.length ft) != (List.length at) then
    sorry (xf "there is %d formal parameters and %d actual parameters in call"
              (List.length ft) (List.length at));
  let var = seq_name "$" in
  let a = fresh_vars (Func {ft_ret_type = TyVar(var);
                            ft_parms = at;
                            ft_tyvars = [] (* Here we cheat a bit... *)
                           })  in
  let a' = match a with Func r -> Func {r with ft_tyvars = [var]} 
                      | _ -> assert false in
  let m = checker c a' (fresh_vars f) in
  try
    let (_, t) = 
      (try
        List.find (fun (x, _) -> x = var) m 
       with Not_found ->
         let (n, _) = List.find (fun (_, x) -> 
                                        match x with 
                                            TyVar(y) when y = var -> true
                                          | _ -> false) m in 
         (n, TyVar n)) in
    match t with
         TyVar k -> 
           if k.[0] == '$' then
             sorry (xf "what the heck? return value of call is '%s" k)
           else
             t
       | _ -> t
  with Not_found ->
    (* hmm... shouldn't we return TyVar(var) here? *)
    sorry "return type of call is unspecified"
;;

let lvalue_p e =
  match e.raw_exp with
      (* Of course this might be wrong... We need some 'const' notation. *)
      Ref r -> ()
    | _ -> sorry "expression is not a proper lvalue"
;;

let rec strip_off ({ctx_tyvars = t} as g) l =
  match l with
      x::xs -> 
        if List.mem x t then
          strip_off g xs
        else
          x :: (strip_off g xs)
    | [] -> []
    
let stripper g t =
  let f t =
    match t with
          Func ({ft_tyvars = tt} as f) -> 
            f.ft_tyvars <- strip_off g tt;
            Func f
        | TyVar _ | NamedType _ | Int | Void | Bool | Float | String 
        | AbstractType | Tuple _  | Struct _ | OptStruct _ 
        | Union _ -> t in
  ignore(type_walk f t)

let rec typecheck_exp (g, e) =
  let subtype_p g t1 t2 =
    try
      subtype g t1 t2; true
    with Type_error _ -> false in
  try
    let (g', t) =
      match e.raw_exp with
          SimpleBinOp(op, e1, e2) ->
            let (g1, t1) = typecheck_exp (g, e1) in
            let (g2, t2) = typecheck_exp (g1, e2) in
            (match op with
                IntPlus | IntMinus | IntDiv | IntMod | IntMult ->
                  subtype g2 Int t1;
                  subtype g2 Int t2;
                  (g2, Int)
              | FloatPlus | FloatMinus | FloatDiv | FloatMult -> 
                  subtype g2 Float t1;
                  subtype g2 Float t2;
                  (g2, Float)
              | Assign ->
                  subtype g2 t1 t2;
                  lvalue_p e1;
                  (g2, t1)
              | Eq | Ne ->
                  if subtype_p g2 Bool t1 then
                    subtype g2 Bool t2
                  else if subtype_p g2 Int t1 then
                    subtype g2 Int t2
                  else if subtype_p g2 Float t1 then
                    subtype g2 Float t2
                  else
                    sorry (xf "`!=' and `==' operators supported only on \
                               bool, int and float types, not on (%s, %s)"
                              (string_of_type t1)
                              (string_of_type t2));
                  (g2, Bool)
              | Lt | Gt | Le | Ge ->
                  if subtype_p g2 Int t1 then
                    subtype g2 Int t2
                  else if subtype_p g2 Float t1 then
                    subtype g2 Float t2
                  else
                    sorry (xf "`<', '>', '<=' and `>=' operators supported \
                               only on int and float types, not on (%s, %s)"
                              (string_of_type t1)
                              (string_of_type t2));
                  (g2, Bool)
              )
        | SimpleUnOp(op, e) ->
            let (g', t) = typecheck_exp (g, e) in
            (match op with
                 IntNegate -> subtype g' Int t; (g', Int)
               | FloatNegate -> subtype g' Float t; (g', Float))
        | Ref {ref_symbol = Some sym} -> (g, sym#get_type)
        | Ref {ref_symbol = None} -> assert false
        | Call(f, act) ->
            let (g', t) = typecheck_exp (g, f) in
            let (g'', tact) = list_apply typecheck_exp (g', act) in
            (g'', typecheck_call g'' t tact)
        | IntConst _ -> (g, Int)
        | BoolConst _ -> (g, Bool)
        | FloatConst _ -> (g, Float)
        | StringConst _ -> (g, String)
        | CrossModuleRef {xmr_symbol = Some sym} -> (g, sym#get_type)
        | CrossModuleRef {xmr_symbol = None} -> assert false
        | Lambda l  ->
            (* Here we need to strip off type variables, that was
             * put here by stupid parser. *)
            List.iter (fun {parm_type=t} -> stripper g t) l.lambda_parms;
            let tyvars = (strip_off g l.lambda_tyvars) @ g.ctx_tyvars in
            ignore(typecheck_stmt ({g with 
                                    ctx_ret_type = l.lambda_ret_type;
                                    ctx_tyvars = tyvars},
                                   l.lambda_body));
            (g, Func {ft_ret_type = l.lambda_ret_type;
                      ft_parms = List.map (fun {parm_type=t} -> t)
                                          l.lambda_parms;
                      ft_tyvars = tyvars}) in
    e.exp_type <- Some t;
    (g', t)
  with Type_error s -> raise (Syntax_error (e.exp_loc, s))

and typecheck_stmt (g, s) =
  try
    match s.raw_stmt with
        Skip -> 
          (g, Void)
      | Exp(e) -> 
          let (g', _) = typecheck_exp (g, e) in 
          (g', Void)
      | Return(e) -> 
        let (g', t) = typecheck_exp (g, e) in
        subtype g g.ctx_ret_type t; 
        (g', Void)
      | ReturnVoid ->
        subtype g g.ctx_ret_type Void;
        (g, Void)
      | Block(ss) ->
        let (g', _) = list_apply typecheck_stmt (g, ss) in
        (g', Void)
      | VarDef {var_type = t} ->
        (* Here we need to strip off type variables, that was
         * put here by stupid parser. *)
        stripper g t;
        (g, Void)
      | Open(_) ->
        (g, Void)
      | Pair(a, b) ->
        let (g', _) = typecheck_stmt (g, a) in
        let (g'', _) = typecheck_stmt (g', b) in
        (g'', Void)
      | BreakIfNot e ->
        let (g', t) = typecheck_exp (g, e) in
        subtype g Bool t;
        (g', Void)
      | If(e, s1, s2) ->
        let (g1, t) = typecheck_exp (g, e) in
        let (g2, _) = typecheck_stmt (g1, s1) in
        let (g3, _) = typecheck_stmt (g2, s2) in
        subtype g Bool t;
        (g3, Void)
      | Loop s ->
        typecheck_stmt (g, s)
  with Type_error e -> raise (Syntax_error (s.stmt_loc, e))
;;

let typecheck_decl d =
  match d with
      FuncDecl f -> 
        ignore(typecheck_stmt ({ctx_ret_type = f.fn_ret_type; 
                                ctx_tyvars = f.fn_tyvars}, f.fn_body))
    | GlobalDecl _ -> ()
    | ExternC _ -> ()
    | FuncProto _ -> ()
    | TopLevelOpen _ -> ()
    | TypeDef _ -> () (* FIXME *)
;;

let type_of e =
  match e.exp_type with
      Some t -> t
    | None -> assert false
;;
